! this module implements an hashtable.
! the size of the table is automatically adjusted when entries are added
! usage is made of Knuth's algoritm for ordered hash tables:
! https://academic.oup.com/comjnl/article-pdf/17/2/135/1405739/170135.pdf
! Scientific American, 1977 april, pp 63-80
!

!MODHASH(3)                 Library Functions Manual                 MODHASH(3)
!
!NAME
!       MODHASH  -  Implementation  of an ordered, automatically resizing, hash
!       table for fortran 90
!
!SYNOPSYS
!       use modhash
!
!       subroutine init(this,m,keylen,notfound): initialize hash table
!       this:  class(hashtable)
!
!       m: integer, intent(in), optional: Size of table, default 100
!       if m < 1 or absent : size will be 100
!
!       keylen: integer, intent(in), optional: Use at least keylen and at  most
!       2*keylen characters computing the hash value.
!       If keylen .le. 0: use all characters.
!
!       notfound:  character(len=*), intent(in), optional: Value to return from
!       lookup() when key is not found
!       default ""
!
!       usetrim: logical, intent(in), optional: .true.: Use trim  on  keys  and
!       values, else do not use trim
!       default .true.
!
!       Note: init can be used on an existing hashtable, the existing hashtable
!       will be cleaned.
!       See also: resize.
!
!       function insert(this,key,val) result(r): insert key/value pair in  hash
!       table
!       this:  class(hashtable)
!
!       key, val: character(len=*), intent(in): Key/value pair to insert in the
!       hash table.
!
!       r:   logical, intent(out): .true. on success
!       Normally, one expects this always to be .true.
!       Note: if the number of items exceeds 0.7 * the size of  the  hashtable,
!       the hashtable is resized.
!
!       function  lookup(this,key,success) result(val): return value associated
!       with key
!       this: class(hashtable)
!
!       key:  character(len=*), intent(in): Key to search for.
!
!       success: logical, intent(out), optional: .true. if key is  found,  else
!       .false.
!
!       val: character(len=:), allocatable: Value associated with key.
!       If key is not found, val = notfound (see init()).
!
!       subroutine delete(this,key): deletes key/value pair
!       this: class(hashtable)
!
!       key:  character(len=*), intent(in): Key to delete.
!
!       Note:  the key/value pair is marked as deleted, if you want to optimize
!       the hash table after many deletions, use resize().
!
!       function find(this,key) result(j): find index of key in hash table
!       this: class(hashtable)
!
!       key:  character(len=*), intent(in): Key to find.
!
!       j:    integer: Index to be used with getkey and getval.
!       If key not found: j = 0
!       Usage is not recommended.
!       See also: lookup()
!
!       function getkey(this,j) result(key): return key at index
!       this: class(hashtable)
!
!       j:    integer, intent(in): Index to use.
!
!       key: character(len=:), allocatable: Key on index j.
!       Note: no check is done on validity of j.
!       Usage is not recommended.
!       See also: lookup()
!
!       function getval(this,j) result(val): return value at index
!       this: class(hashtable)
!
!       j:    integer, intent(in): Index to use.
!
!       val: character(len=:), allocatable: Value on index j.
!       Note: No check is done on validity of j.
!       Usage is not recommended.
!       See also: lookup()
!
!       subroutine stats(this,m,items,collisions,resizes,deletions,
!       keylen,notfound): statistics about the hash table
!       this: class(hashtable)
!
!       m:    integer, intent(out), optional: Size of hash table.
!
!       items: integer, intent(out), optional: Number  of  key/value  pairs  in
!       hash table, including deleted items.
!
!       collisions:  integer, intent(out), optional: Total number of collisions
!       (i.e. different keys with same hash value).
!
!       resizes: integer, intent(out), optional: Total number of resizes of the
!       hash table.
!
!       deletions:  integer,  intent(out),  optional:  Total  number of deleted
!       items in the hash table.
!
!       keylen: integer, intent(out), optional: Keylen used, see init().
!
!       notfound: character(len=:), intent(out), allocatable,  optional:  Value
!       returned by lookup() if key not found.
!
!       subroutine  getentry(this,  n, r, key, val, del): main purpose: Get all
!       entries from the hash table.
!       this: class(hashtable)
!
!       n:     integer, intent(inout)
!       input: index of item to visit
!       output: actual index used
!
!       r:     logical, intent(out): .true. if item found, else .false.
!
!       del:   logical, intent(out): .true. if found item was deleted,  .false.
!       otherwize
!
!       key, val: character(len=:), allocatable, intent(out): key/value pair of
!       item.  If not found: val = notfound (see init())
!
!       Note: this subroutine can be used to list all entries in the  hash  ta‐
!       ble, see example.
!
!       subroutine resize(this,n): resize hash table, get rid of deletions
!       this: class(hashtable)
!
!       n:    integer, intent(in), optional
!       if  not  present: New size is calculated to allocate space for approxi‐
!       mately two times the current not-deleted items.
!       == 0: New size is calculated to hold the current not-deleted items.
!       != 0: The new size will be n. If this value is to low to hold the  not-
!       deleted items: no resizing will be done.
!
!       Note: resizing is automatically done in insert() when needed.
!
!       subroutine cleanup(this) : deallocates contents of hash table
!       this: class(hashtable)
!
!EXAMPLE PROGRAM
!        ---->
!
!        program example
!          use modhash
!          implicit none
!          type(hashtable) :: h
!          logical         :: success, del
!          integer         :: j,dels,items,i
!          character(len=:), allocatable :: key,val
!          print *,"Initializing..."
!          call h%init(notfound="NOT FOUND")
!          print *,"Inserting..."
!          success = h%insert("first key","first value")
!          success = h%insert("2nd key","2nd value")
!          success = h%insert("3rd key","3rd value")
!          j = h%find("first key")   ! not recommended, use lookup
!          key = h%getkey(j)         ! not recommended, use lookup
!          val = h%getval(j)         ! not recommended, use lookup
!          print *, "Using stats ..."
!          call h%stats(deletions=dels, items=items)
!          print *,"deletions:",dels,"items",items
!          print *,"Using lookup ..."
!          val = h%lookup("2nd key")
!          print *,"2nd key: ",val
!          print *,"Using lookup with not existing key ..."
!          val = h%lookup("xnd key")
!          print *,"xnd key: ",val
!          print *,"Deleting:"
!          call h%delete("first key")
!          print *, "Using stats again ..."
!          call h%stats(deletions=dels, items=items)
!          print *,"deletions:",dels,"items",items
!          print *,"All entries ..."
!          i = 1
!          success = .true.
!          do while(success)
!             call h%getentry(i, success, key, val, del)
!             if (success) print *,"key: ",key," value: ",val," deleted:",del
!             i = i+1
!          enddo
!          print *,"Resizing ..."
!          call h%resize(0)
!          print *, "Using stats again ..."
!          call h%stats(deletions=dels, items=items)
!          print *,"deletions:",dels,"items",items
!          print *,"removing contents of hash table ..."
!          call h%cleanup
!        end program example
!
!        <----
!
!EXAMPLE PROGRAM USING TRANSFER
!        ---->
!
!       !  example  using modhash with other data types, using transfer program
!       cryst
!          use modhash
!          implicit none
!          integer(selected_int_kind(3)) :: hkl(3,2)
!          type reftype
!             real e
!             real r
!             integer(selected_int_kind(2)) :: m
!          end type reftype
!          type(reftype)   :: a,b,c
!          type(hashtable) :: h
!          real            :: eps = 0.00001
!
!          character(len=:), allocatable :: hc,rc
!          logical                       :: success
!
!          allocate(character(len=sizeof(hkl(:,1))) :: hc)
!          allocate(character(len=sizeof(a))        :: rc)
!
!          print *,"len hc:",len(hc)
!          print *,"len rc:",len(rc)
!          hkl(1,1) = 2
!          hkl(2,1) = 3
!          hkl(3,1) = 0
!          a%e=2.1
!          a%m=4
!          a%r=0.3
!
!          hkl(1,2) = 0
!          hkl(2,2) = 7
!          hkl(3,2) = 3
!
!          b%e=1.8
!          b%m=1
!          b%r=0.2
!
!          call h%init(usetrim=.false.)
!
!          success = h%insert(transfer(hkl(:,1),hc),transfer(a,rc))
!          success = h%insert(transfer(hkl(:,2),hc),transfer(b,rc))
!
!          c = transfer(h%lookup(transfer(hkl(:,1),hc)),c)
!          print *,c
!          if (aboutequal(c%e, a%e) .and. &
!          &   aboutequal(c%r, a%r) .and. &
!          &   c%m == a%m) then
!             print *,'OK'
!          else
!             stop 1
!          endif
!          c = transfer(h%lookup(transfer(hkl(:,2),hc)),c)
!          print *,c
!          if (aboutequal(c%e, b%e) .and. &
!          &   aboutequal(c%r, b%r) .and. &
!          &   c%m == b%m) then
!             print *,'OK'
!          else
!             stop 1
!          endif
!          c = transfer(h%lookup(transfer([4,5,6],hc),success),c)
!          if (success) then
!             print *,c
!             stop 1
!          else
!             print *,"not found"
!             print *,'OK'
!          endif
!
!          print *,'all is well'
!
!          contains
!          logical function aboutequal(x,y)
!             implicit none
!             real, intent(in) :: x,y
!             real             :: eps = 0.0001
!             aboutequal = (x > y - eps .and. x < y + eps)
!          end function aboutequal
!
!       end program cryst
!
!        <----
!
!       Compile this programs like:
!
!       If you installed modhash 'properly' and have pkgconfig:
!
!       gfortran `pkg-config --cflags modhash` -o example example.f90 \
!                `pkg-config --libs modhash`
!
!       Otherwise if modhash is installed in /home/me/software:
!
!       gfortran -I/home/me/include -o example example.f90 \
!                -L/home/me/lib -lmodhash
!
!       
!       COPYRIGHT
!       This is free software; see the source for copying conditions.  There is NO
!       warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
!
!                                     2018                           MODHASH(3)

module modhash
   implicit none
   private

   type htype
      integer(selected_int_kind(1)) :: flags ! iand(flags,1): present, iand(flags,2): deleted
      !                                      ! see ispresent() and isdeleted() below
      character(len = :),allocatable  :: key
      character(len = :),allocatable  :: val
   end type htype

   real, parameter :: growcrit   = 0.7          ! used by insert(). Grow hash table when number of items > growcrit*nhash

   type, public :: hashtable
      private
      integer :: nhash        ! size of the hash table
      integer :: collisions   ! number of collisions during insert
      integer :: curtable     ! 1: actual hashtable is htable1, 2: actual hashtable = htable2
      integer :: nresize      ! number of resizes of hash table
      integer :: deletions    ! number of deleted items in hash table
      integer :: nitems       ! number of items in hash table, including deleted itemd
      integer :: keylen       ! approximately number of characters to use for computing hash
      logical :: usetrim      ! use (.true.) or not (.false.) use trim function on strings
      ! in order to be able to resize, we need two hash tables:
      type(htype), pointer :: htable1(:) => Null()
      type(htype), pointer :: htable2(:) => Null()
      type(htype), pointer :: whtable(:) => Null() ! whtable will point to the current hash table
      character(len = :), allocatable :: notfound  ! used when key is not found i e.g. lookup()

   contains

      procedure, public  :: cleanup  => hash_cleanup
      procedure, public  :: delete   => hash_delete
      procedure, public  :: find     => hash_find
      procedure, public  :: getentry => hash_getentry
      procedure, public  :: getkey   => hash_getkey
      procedure, public  :: getval   => hash_getval
      procedure, public  :: init     => hash_init
      procedure, public  :: insert   => hash_insert
      procedure, public  :: lookup   => hash_lookup
      procedure, public  :: resize   => hash_resize
      procedure, public  :: stats    => hash_stats

      procedure, private :: hash     => hash_function
      procedure, private :: insert_1 => hash_insert_1

   end type hashtable

contains
! this computes a hash from a string. See https://en.wikipedia.org/wiki/Jenkins_hash_function
! Not exactly what the original code prescribes, but it seems to work well.

   function jenkins_one_at_a_time_hash(key,keylen,usetrim) result (hash)

      implicit none
      character (len = *), intent(in) :: key      ! string for which hash is to be computed
      integer            , intent(in) :: keylen   ! for long strings, it can be advantageous to use
      !                                           ! not all characters. Using keylen > 0, the number
      !                                           ! of characters used is keylen .. 2* keylen
      logical            , intent(in) :: usetrim
      integer                         :: hash

      integer :: i, d, jx, m

      hash = 0
      if (usetrim) then
         m = len_trim(key)
      else
         m = len(key)
      endif
      if (keylen .le. 0) then
         do i = 1,m
            hash = hash + iachar(key(i:i))
            hash = hash + ishft(hash,10)
            hash = ieor(hash,shifta(hash,6))
         enddo
      else
         d  = max(256,256*m/keylen)
         jx = 256
         i  = 1
         do while(i .le. m)
            hash = hash + iachar(key(i:i))
            hash = hash + ishft(hash,10)
            hash = ieor(hash,shifta(hash,6))
            jx = jx + d
            i = jx/256
         enddo
      endif
      hash = hash + ishft(hash,3)
      hash = ieor(hash,shifta(hash,11))
      hash = hash + ishft(hash,15)
   end function jenkins_one_at_a_time_hash

!       static unsigned long
!    sdbm(str)
!    unsigned char *str;
!    {
!        unsigned long hash = 0;
!        int c;
!
!        while (c = *str++)
!            hash = c + (hash << 6) + (hash << 16) - hash;
!
!        return hash;
!        }
   function sdbm(key,keylen) result(hash)
      implicit none
      character(len = *), intent(in) :: key
      integer,            intent(in) :: keylen
      integer                        :: hash

      integer :: i,imax
      hash = 0
      if (keylen <= 0) then
         imax = len(key)
      else
         imax = min(len(key),keylen)
      endif

      do i=1,imax
         hash = ichar(key(i:i)) + shiftl(hash,6) + shiftl(hash,16) - hash;
      enddo

   end function sdbm


   function hash_function(this,key) result(h)
      ! this function takes care that the result of the jenkins function is between 1 and nhash
      implicit none
      class(hashtable)               :: this
      character(len = *), intent(in) :: key
      integer                        :: h

      h = mod(abs(jenkins_one_at_a_time_hash(key,this%keylen,this%usetrim)),this%nhash) + 1
      !h = mod(abs(sdbm(key,this%keylen)),this%nhash) + 1
   end function hash_function


   subroutine hash_init(this,m,keylen,notfound,usetrim)
      ! see above
      implicit none
      class (hashtable)                        :: this
      integer,            intent(in), optional :: m, keylen
      character(len = *), intent(in), optional :: notfound
      logical,            intent(in), optional :: usetrim

      integer i

      this%nitems     = 0
      this%nresize    = 0
      this%collisions = 0
      this%deletions  = 0
      this%curtable   = 1

      if (present(m)) then
         this%nhash = m
      else
         this%nhash = 100
      endif

      if (this%nhash .le. 1) then
         this%nhash = 100
      endif

      this%nhash = fakeprimege(this%nhash)

      if (present(keylen)) then
         this%keylen = keylen
      else
         this%keylen = 0
      endif

      if (present(notfound)) then
         this%notfound = notfound
      else
         this%notfound = ""
      endif

      if (present(usetrim)) then
         this%usetrim = usetrim
      else
         this%usetrim = .true.
      endif

      call this%cleanup
      allocate (this%htable1(this%nhash))
      this%whtable => this%htable1
      do i = 1,this%nhash
         this%whtable(i)%flags = 0  ! mark all entries as not-present and not-deleted
      enddo
   end subroutine hash_init

   subroutine hash_resize(this,n)
      ! see above
      implicit none
      class(hashtable)              :: this
      integer, intent(in), optional :: n

      integer              :: nnhash,i,imax
      type(htype), pointer :: tmphtable(:)
      logical              :: r


      if (present(n)) then
         if (n .eq. 0) then
            nnhash = int((this%nitems - this%deletions)/growcrit+1.0)
         else if (n .lt. this%nitems-this%deletions) then
            return
         else
            nnhash = n
         endif
      else
         nnhash = int(1+2.0/growcrit*(this%nitems - this%deletions))
      endif

      nnhash = fakeprimege(max(nnhash,2))

      this%nresize  = this%nresize+1

      ! let tmphtable point to current hash table
      ! allocate new hash table, and let whtable point to it

      if (this%curtable .eq. 1) then
         allocate(this%htable2(nnhash))
         tmphtable     => this%htable1
         this%whtable  => this%htable2
         this%curtable = 2
      else
         allocate(this%htable1(nnhash))
         tmphtable     => this%htable2
         this%whtable  => this%htable1
         this%curtable = 1
      endif

      do i = 1,nnhash
         this%whtable(i)%flags = 0  ! all entries are marked as not-present and not-deleted
      enddo

      ! fill new hash table

      this%nitems     = 0
      imax            = this%nhash
      this%nhash      = nnhash
      this%collisions = 0
      this%deletions  = 0
      do i = 1,imax
         if (ispresent(tmphtable(i))) then
            if (.not. isdeleted(tmphtable(i))) then
               r = this%insert_1(tmphtable(i)%key,tmphtable(i)%val)
               if ( .not. r) then
                  print *,"Error when resizing: not enough space"
                  stop
               endif
            endif
         endif
      enddo

      ! cleanup old hash table

      if (this%curtable .eq. 1) then
         call cleanup_table(this%htable2)
      else
         call cleanup_table(this%htable1)
      endif
   end subroutine hash_resize

   subroutine hash_cleanup(this)
      implicit none
      class(hashtable) :: this

      call cleanup_table(this%htable1)
      call cleanup_table(this%htable2)
   end subroutine hash_cleanup

   function hash_find(this,key) result(j)
      ! see above
      implicit none
      class(hashtable)               :: this
      character(len = *), intent(in) :: key
      integer                        :: j

      integer :: l

      j = this%hash(key)
      do l = 1,this%nhash
         if (.not. ispresent(this%whtable(j))) then
            j = 0
            return
         endif
         if (this%whtable(j)%key .lt. key) then
            j = 0
            return
         endif
         if (this%whtable(j)%key .eq. key) then
            if (isdeleted(this%whtable(j))) then
               j = 0
               return
            endif
            return
         endif
         j = j-1
         if (j .eq. 0) then
            j = this%nhash
         endif
      enddo
      j = 0
   end function hash_find

   function hash_insert(this,key, val) result(r)
      ! see above
      implicit none
      class(hashtable)               :: this
      character(len = *), intent(in) :: key, val
      logical                        :: r

      if (this%nitems .gt. int(growcrit*this%nhash) &
         .or. this%nitems .ge. this%nhash-1) then
         call this%resize
      endif
      r = this%insert_1(key,val)
   end function hash_insert

   function hash_insert_1(this,key, val) result(r)
      ! see above
      implicit none
      class(hashtable)               :: this
      character(len = *), intent(in) :: key, val
      logical                        :: r

      integer     :: j,l
      type(htype) :: d,x

      x%flags = 1
      if (this%usetrim) then
         x%key   = trim(key)
         x%val   = trim(val)
      else
         x%key   = key
         x%val   = val
      endif

      j = this%hash(key)
      do l = 1,this%nhash
         if (ispresent(this%whtable(j))) then
            if (this%whtable(j)%key .lt. x%key) then
               d          = this%whtable(j)
               this%whtable(j) = x
               x          = d
               this%collisions = this%collisions + 1
            else if (isdeleted(this%whtable(j))) then
               this%whtable(j) = x
               r               = .true.
               return
            else if (this%whtable(j)%key .eq. x%key) then
               this%whtable(j) = x
               r               = .true.
               return
            endif
         else
            this%whtable(j) = x
            this%nitems = this%nitems + 1
            r = .true.
            return
         endif
         j = j-1
         if (j .eq. 0) then
            j = this%nhash
         endif
      enddo
      r = .false.
   end function hash_insert_1

   function hash_getkey(this,j) result(key)
      ! see above
      implicit none
      class(hashtable)                :: this
      integer, intent(in)             :: j
      character(len = :), allocatable :: key

      key = this%whtable(j)%key
   end function hash_getkey

   function hash_getval(this,j) result(val)
      ! see above
      implicit none
      class(hashtable)                :: this
      integer, intent(in)             :: j
      character(len = :), allocatable :: val

      val = this%whtable(j)%val
   end function hash_getval

   subroutine hash_stats(this,m,items,collisions,resizes,deletions,keylen,notfound,usetrim)
      ! see above
      implicit none
      class(hashtable)                                       :: this
      integer, intent(out), optional                         :: m,items,collisions,resizes,deletions,keylen
      character(len = :), intent(out), allocatable, optional :: notfound
      logical, intent(out), optional                         :: usetrim

      if (present(m))          m          = this%nhash
      if (present(collisions)) collisions = this%collisions
      if (present(resizes))    resizes    = this%nresize
      if (present(items))      items      = this%nitems
      if (present(deletions))  deletions  = this%deletions
      if (present(keylen))     keylen     = this%keylen
      if (present(notfound))   notfound   = this%notfound
      if (present(usetrim))    usetrim    = this%usetrim
   end subroutine hash_stats

   subroutine hash_getentry(this, n, r, key, val, del)
      ! see above
      implicit none
      class(hashtable)                             :: this
      integer, intent(inout)                       :: n
      logical, intent(out)                         :: r,del
      character(len = :), allocatable, intent(out) :: key,val

      if (n .gt. this%nhash) then
         key = ""
         val = this%notfound
         r   = .false.
         del = .false.
         return
      endif

      do
         if (ispresent(this%whtable(n))) then
            key = this%whtable(n)%key
            val = this%whtable(n)%val
            del = isdeleted(this%whtable(n))
            r   = .true.
            return
         endif
         n = n+1
         if (n .gt. this%nhash) then
            key = ""
            val = this%notfound
            r   = .false.
            del = .false.
            return
         endif
      enddo
   end subroutine hash_getentry

   subroutine hash_delete(this,key)
      ! see above
      implicit none
      class(hashtable)               :: this
      character(len = *), intent(in) :: key

      integer :: j

      j = this%find(key)
      if (j .ne. 0) then
         if (.not. isdeleted(this%whtable(j))) then
            this%deletions = this%deletions + 1
         endif
         this%whtable(j)%flags = 3
      endif
   end subroutine hash_delete

   function hash_lookup(this,key,success) result(val)
      ! see above
      implicit none
      class(hashtable)                 :: this
      character(len = *), intent(in)   :: key
      logical, intent(out), optional   :: success
      character(len = :), allocatable  :: val

      integer :: j

      j = this%find(key)
      if (j .eq. 0) then
         val = this%notfound
         if (present(success)) then
            success = .false.
         endif
      else
         val = this%getval(j)
         if (present(success)) then
            success = .true.
         endif
      endif
   end function hash_lookup

   function fakeprimege(n) result(nge)
      ! this function is a poor mans implementation of 'find smallest prime >= n'
      ! only a few primes are used to check divisionability
      ! this function is here, because the size of an hash table should be
      ! a prime number
      implicit none
      integer, intent(in)  :: n
      integer              :: nge

      integer, parameter   :: np = 10
      integer, parameter   :: primes(np) = (/2,3,5,7,11,13,17,19,23,29/)
      integer              :: i
      logical              :: isprime

      nge = max(abs(n),2)                     ! guard against strange input
      nge = nge + 1 - mod(nge,2)              ! make nge uneven

      do
         isprime = .true.
         do i = 1,np
            if (nge == primes(i)) return      ! nge is one of primes
            if (mod(nge,primes(i)) == 0) then ! nge divisible by one of primes
               isprime = .false.
               exit
            endif
         enddo
         if (isprime) return
         nge = nge + 2                        ! try incremented nge
      enddo

   end function fakeprimege

   function ispresent(x) result(r)
      implicit none
      type(htype), intent(in) :: x
      logical                 :: r

      r = iand(x%flags,1) .ne. 0
   end function ispresent

   function isdeleted(x) result(r)
      implicit none
      type(htype), intent(in) :: x
      logical                 :: r

      r = iand(x%flags,2) .ne. 0
   end function isdeleted

   subroutine cleanup_table(t)
      ! cleans hash table t
      implicit none
      type(htype), intent(inout), pointer :: t(:)

      integer :: i

      if (associated(t)) then
         do i = 1,size(t)
            if (ispresent(t(i))) then
               deallocate(t(i)%key,t(i)%val)
            endif
         enddo

         deallocate(t)
      endif
   end subroutine cleanup_table

end module modhash
